using System;
using System.Collections.Generic;
using System.IO;
using System.Windows.Forms;
using System.Xml.Linq;

namespace DAWalkerWorldMapEditor {
    public partial class DAWalkerWorldMapEditorForm : Form {

        private struct Node {
            public struct infoMap {
                private int[][] currInfoMap;

                public infoMap(int[][] infoMap) { this.currInfoMap = infoMap; }
                public int this[int nodes, int index] { get { return currInfoMap[nodes][index]; } }
                public int[] this[int nodes] { get { return currInfoMap[nodes]; } }
                public override string ToString() {
                    string mapPortalLocs = "{";
                    for(int i = 0; i < currInfoMap[1].Length; i++) mapPortalLocs += currInfoMap[1][i].ToString() + " ";
                    mapPortalLocs = mapPortalLocs.TrimEnd(' ') + "}";

                    string specialOps = "{";
                    for(int i = 0; i < currInfoMap[2].Length; i++) specialOps += currInfoMap[2][i].ToString() + " ";
                    specialOps = specialOps.TrimEnd(' ') + "}";

                    string mapName = "";
                    if(mapNodes.ContainsKey(currInfoMap[0][0]) && mapNodes[currInfoMap[0][0]].name != "") mapName = mapNodes[currInfoMap[0][0]].name + " ";

                    return mapName + "(" + currInfoMap[0][0].ToString() + ") " + mapPortalLocs + " " + specialOps;
                }
            }
            public int mapNum;
            public int[] mapAxis;
            public string name;
            public bool otherList;
            public SortedList<int, infoMap> infoPortals;

            public Node(int mapNum, int[] mapAxis, bool otherList) {
                this.mapNum = mapNum;
                this.mapAxis = mapAxis;
                this.name = "";
                this.otherList = otherList;

                infoPortals = new SortedList<int, infoMap>();
            }
            public Node(int mapNum, int[] mapAxis, string name, bool otherList) {
                this.mapNum = mapNum;
                this.mapAxis = mapAxis;
                this.name = name;
                this.otherList = otherList;

                infoPortals = new SortedList<int, infoMap>();
            }
            public void addExit(int destinationMap, int[] mapPortalLocs, int[] specialOps) { infoPortals[destinationMap] = new infoMap(new int[][] { new int[] { destinationMap }, mapPortalLocs, specialOps }); }
            public void removeExit(int infoPortal) { infoPortals.RemoveAt(infoPortal); }
            public int[] getWarpPortalLocs(int destinationMap) {
                foreach(KeyValuePair<int, infoMap> A in infoPortals) if(A.Value[0, 0] == destinationMap) return A.Value[1];
                return null;
            }
            public int[] getSpecialOps(int destinationMap) {
                foreach(KeyValuePair<int, infoMap> A in infoPortals) if(A.Value[0, 0] == destinationMap) return A.Value[2];
                return null;
            }
            public override string ToString() {
                if(name == "") return "(" + mapNum.ToString() + ")";
                else return name + " (" + mapNum.ToString() + ")";
            }
        }

        private static SortedList<int, Node> mapNodes = new SortedList<int, Node>();
        private MapPreview MapPreviewForm;
        private MouseLocator ShowMouseForm;
        private int DAMapAddr;
        private int DAMapAddrOff;


        public DAWalkerWorldMapEditorForm() {
            InitializeComponent();
            this.Text = "World Map Editor v" + Application.ProductVersion;

            if(!System.IO.File.Exists("DAData.xml")) { MessageBox.Show("DAData.xml file not found: Download the file from www.da-wizard.com."); Environment.Exit(0); }
            TextBoxMapFolder.Text = XElement.Load("DAData.xml").Element("Data").Element("MapFolder").Value;
            DAMapAddr = Int32.Parse(XElement.Load("DAData.xml").Element("Pointers").Element("MapNum").Value, System.Globalization.NumberStyles.HexNumber);
            DAMapAddrOff = Int32.Parse(XElement.Load("DAData.xml").Element("Pointers").Element("MapNum").Attribute("o1").Value, System.Globalization.NumberStyles.HexNumber);
        }

        private void ClearWorld() {
            mapNodes = new SortedList<int, Node>();
            ListBoxCurrPortals.DataSource = null;
            ListBoxCurrPortals.Items.Clear();
            ListBoxWorldMaps.DataSource = null;
            ListBoxWorldMaps.Items.Clear();
            TextBoxMapAxis.Clear();
            TextBoxMapNum.Clear();
            TextBoxName.Clear();
            TextBoxSpecialEndOps.Clear();
            TextBoxWarpPortalLocs.Clear();
            TextBoxWarpsTo.Clear();
        }

        private void OpenWorldMap(string filePath) {
            StreamReader worldMapFile;
            try { worldMapFile = File.OpenText(@filePath); } catch(Exception e) {
                MessageBox.Show("Couldn't Load File" + e.Message);
                return;
            }

            string[] currLine;
            while(!worldMapFile.EndOfStream) {
                currLine = worldMapFile.ReadLine().Split(new char[] { ' ' }, StringSplitOptions.RemoveEmptyEntries);

                int mapNum = 0;
                if(currLine[0] != "-1") {
                    mapNum = int.Parse(currLine[0]);
                    if(currLine.Length == 3) mapNodes[mapNum] = new Node(mapNum, new int[] { int.Parse(currLine[1]), int.Parse(currLine[2]) }, false);
                    else {
                        string name = "";
                        for(int i = 3; i < currLine.Length; i++) name += currLine[i] + " ";
                        mapNodes[mapNum] = new Node(mapNum, new int[] { int.Parse(currLine[1]), int.Parse(currLine[2]) }, name.TrimEnd(' '), false);
                    }
                } else {
                    mapNum = int.Parse(currLine[1]);
                    if(currLine.Length == 4) mapNodes[mapNum] = new Node(mapNum, new int[] { int.Parse(currLine[2]), int.Parse(currLine[3]) }, true);
                    else {
                        string name = "";
                        for(int i = 4; i < currLine.Length; i++) name += currLine[i] + " ";
                        mapNodes[mapNum] = new Node(mapNum, new int[] { int.Parse(currLine[2]), int.Parse(currLine[3]) }, name.TrimEnd(' '), true);
                    }
                }

                if(!worldMapFile.EndOfStream) currLine = worldMapFile.ReadLine().Split(new char[] { ' ' }, StringSplitOptions.RemoveEmptyEntries);
                while(!worldMapFile.EndOfStream && currLine.Length != 0) {
                    int destinationMap = int.Parse(currLine[0]);

                    int[] specialOps = new int[currLine.Length - 1];
                    for(int i = 1; i < currLine.Length; i++) specialOps[i - 1] = int.Parse(currLine[i]);

                    currLine = worldMapFile.ReadLine().Split(new char[] { ' ' }, StringSplitOptions.RemoveEmptyEntries);

                    int[] mapPortalLocs = new int[currLine.Length];
                    for(int i = 0; i < currLine.Length; i++) mapPortalLocs[i] = int.Parse(currLine[i]);

                    mapNodes[mapNum].addExit(destinationMap, mapPortalLocs, specialOps);
                    if(!worldMapFile.EndOfStream) currLine = worldMapFile.ReadLine().Split(new char[] { ' ' }, StringSplitOptions.RemoveEmptyEntries);
                }
            }
            worldMapFile.Close();

            ListBoxWorldMaps.DataSource = new List<Node>(mapNodes.Values);
        }
        private void ButtonLoadMapFile_Click(object sender, EventArgs e) {
            OpenFileDialog DAWorldMapFileBrowse = new OpenFileDialog();
            DAWorldMapFileBrowse.Title = "Open DAWalker World Map File";
            DAWorldMapFileBrowse.Filter = "DAWorld Map File (*.dat)|*.dat";
            DAWorldMapFileBrowse.InitialDirectory = Application.StartupPath;
            DAWorldMapFileBrowse.RestoreDirectory = true;
            if(DAWorldMapFileBrowse.ShowDialog() == DialogResult.OK) {
                ClearWorld();
                OpenWorldMap(DAWorldMapFileBrowse.FileName);
            }
        }

        private void ButtonSaveWorldMap_Click(object sender, EventArgs e) {
            SaveFileDialog DAWorldMapFileBrowse = new SaveFileDialog();
            DAWorldMapFileBrowse.Title = "Save DAWalker World Map File";
            DAWorldMapFileBrowse.Filter = "DAWorld Map File (*.dat)|*.dat";
            DAWorldMapFileBrowse.InitialDirectory = Application.StartupPath;
            DAWorldMapFileBrowse.RestoreDirectory = true;
            if(DAWorldMapFileBrowse.ShowDialog() == DialogResult.OK) {
                StreamWriter worldMapFile;
                try { worldMapFile = File.CreateText(@DAWorldMapFileBrowse.FileName); } catch(Exception eX) {
                    MessageBox.Show("Couldn't Load File" + eX.Message);
                    return;
                }

                XElement myXML = XElement.Load("DAData.xml");
                myXML.Element("Data").Element("MapFolder").SetValue(TextBoxMapFolder.Text);
                myXML.Save("DAData.xml");

                bool first = true;
                foreach(KeyValuePair<int, Node> A in mapNodes) {
                    if(!first) worldMapFile.WriteLine();
                    first = false;

                    if(A.Value.otherList) worldMapFile.Write("-1 ");
                    if(A.Value.name != "") worldMapFile.WriteLine(A.Value.mapNum + " " + A.Value.mapAxis[0] + " " + A.Value.mapAxis[1] + " " + A.Value.name);
                    else worldMapFile.WriteLine(A.Value.mapNum + " " + A.Value.mapAxis[0] + " " + A.Value.mapAxis[1]);

                    foreach(KeyValuePair<int, Node.infoMap> B in A.Value.infoPortals) {
                        string writeStringOpsPortals = " ";
                        foreach(int C in B.Value[2]) writeStringOpsPortals += C + " ";
                        worldMapFile.WriteLine(B.Value[0, 0] + writeStringOpsPortals.TrimEnd(' '));

                        writeStringOpsPortals = "";
                        foreach(int C in B.Value[1]) writeStringOpsPortals += C + " ";
                        worldMapFile.WriteLine(writeStringOpsPortals.TrimEnd(' '));
                    }
                }
                worldMapFile.Close();
            }
        }

        private void ButtonAddMap_Click(object sender, EventArgs e) {
            string[] axis = TextBoxMapAxis.Text.Split(new char[] { ' ' }, StringSplitOptions.RemoveEmptyEntries);
            if(axis.Length != 2 || ListBoxWorldMaps.SelectedIndex == -1 || TextBoxMapNum.Text == "") {
                MessageBox.Show("Dimensions arn't correct, not Filled In, or Not Avaliable");
                return;
            }

            int index = int.Parse(TextBoxMapNum.Text);
            if(!mapNodes.ContainsKey(index)) {
                mapNodes[index] = new Node(index, new int[] { int.Parse(axis[0]), int.Parse(axis[1]) }, TextBoxName.Text, CheckBoxOtherList.Checked);

                ListBoxWorldMaps.DataSource = new List<Node>(mapNodes.Values);
                ListBoxWorldMaps.SelectedItem = mapNodes[index];
            } else if(index == ((Node)ListBoxWorldMaps.SelectedItem).mapNum) {
                string[] mapAxis = TextBoxMapAxis.Text.Split(new char[] { ' ' }, StringSplitOptions.RemoveEmptyEntries);
                if(mapAxis.Length != 2) {
                    MessageBox.Show("Axis not in Correct Formate");
                    return;
                }

                Node currNode = mapNodes[index];
                currNode.mapAxis = new int[] { int.Parse(mapAxis[0]), int.Parse(mapAxis[1]) };
                currNode.name = TextBoxName.Text;
                mapNodes[index] = currNode;

                ListBoxWorldMaps.DataSource = new List<Node>(mapNodes.Values);
                ListBoxWorldMaps.SelectedItem = currNode;

                if(MapPreviewForm != null && MapPreviewForm.Visible) {
                    MapPreviewForm.OpenDAMapFile();
                    ListBoxWorldMaps_SelectedIndexChanged(null, null);
                }
            } else {
                MessageBox.Show("Map Already Exists, Switching To");
                ListBoxWorldMaps.SelectedItem = mapNodes[index];
            }
        }

        private void ButtonOpenMapViewer_Click(object sender, EventArgs e) {
            if(MapPreviewForm == null) MapPreviewForm = new MapPreview(this);
            MapPreviewForm.Show();
            MapPreviewForm.OpenDAMapFile();

            ListBoxWorldMaps_SelectedIndexChanged(null, null); //Updates the portals on the View Map
        }

        private void BTNAddPortal_Click(object sender, EventArgs e) {
            if(ListBoxWorldMaps.SelectedIndex == -1 || TextBoxWarpsTo.Text == "") return;

            string[] mapPortalLocsS = TextBoxWarpPortalLocs.Text.Split(new char[] { ' ' }, StringSplitOptions.RemoveEmptyEntries);
            int[] mapPortalLocs = new int[mapPortalLocsS.Length];
            for(int i = 0; i < mapPortalLocsS.Length; i++) mapPortalLocs[i] = int.Parse(mapPortalLocsS[i]);

            string[] specialOpsS = TextBoxSpecialEndOps.Text.Split(new char[] { ' ' }, StringSplitOptions.RemoveEmptyEntries);
            int[] specialOps = new int[specialOpsS.Length];
            for(int i = 0; i < specialOpsS.Length; i++) specialOps[i] = int.Parse(specialOpsS[i]);

            int index = ((Node)ListBoxWorldMaps.SelectedItem).mapNum;
            mapNodes[index].addExit(int.Parse(TextBoxWarpsTo.Text), mapPortalLocs, specialOps);
            ListBoxCurrPortals.DataSource = new List<Node.infoMap>(mapNodes[index].infoPortals.Values);
        }

        private void btnMouseCoords_Click(object sender, EventArgs e) {
            ShowMouseForm = new MouseLocator();
            ShowMouseForm.MainF = this;
            ShowMouseForm.Show();
            btnMouseCoords.Enabled = false;
        }

        private void ListBoxWorldMaps_SelectedIndexChanged(object sender, EventArgs e) {
            if(ListBoxWorldMaps.DataSource == null) return;

            int index = ((Node)ListBoxWorldMaps.SelectedItem).mapNum;
            CheckBoxOtherList.Checked = mapNodes[index].otherList;
            TextBoxName.Text = mapNodes[index].name;
            TextBoxMapNum.Text = mapNodes[index].mapNum.ToString();
            TextBoxMapAxis.Text = mapNodes[index].mapAxis[0] + " " + mapNodes[index].mapAxis[1];
            ListBoxCurrPortals.DataSource = new List<Node.infoMap>(mapNodes[index].infoPortals.Values);

            if(MapPreviewForm != null && MapPreviewForm.Visible && ListBoxCurrPortals.SelectedIndex == -1) MapPreviewForm.OpenDAMapFile();
        }

        private void ListBoxCurrPortals_SelectedIndexChanged(object sender, EventArgs e) {
            if(ListBoxCurrPortals.DataSource == null || ((ListBoxWorldMaps.SelectedIndex != -1) && (((Node)ListBoxWorldMaps.SelectedItem).mapNum.ToString() != TextBoxMapNum.Text))) return;
            Node.infoMap currInfoMap = ((Node.infoMap)ListBoxCurrPortals.SelectedItem);

            string mapPortalLocs = "";
            for(int i = 0; i < currInfoMap[1].Length; i++) mapPortalLocs += currInfoMap[1][i].ToString() + " ";

            string specialOps = "";
            for(int i = 0; i < currInfoMap[2].Length; i++) specialOps += currInfoMap[2][i].ToString() + " ";

            TextBoxWarpPortalLocs.Text = mapPortalLocs.TrimEnd(' ');
            TextBoxWarpsTo.Text = currInfoMap[0][0].ToString();
            TextBoxSpecialEndOps.Text = specialOps.TrimEnd(' ');

            if(MapPreviewForm != null && MapPreviewForm.Visible && ListBoxCurrPortals.DataSource != null) {
                MapPreviewForm.OpenDAMapFile();
                MapPreviewForm.DrawSquareBlockAndUpdate(((Node.infoMap)ListBoxCurrPortals.SelectedItem)[1]);
            }
        }

        private void LabelDelPortal_Click(object sender, EventArgs e) {
            if(ListBoxWorldMaps.SelectedIndex == -1 || ListBoxCurrPortals.SelectedIndex == -1) return;
            int index = ((Node)ListBoxWorldMaps.SelectedItem).mapNum;
            mapNodes[index].removeExit(ListBoxCurrPortals.SelectedIndex);
            ListBoxCurrPortals.DataSource = new List<Node.infoMap>(mapNodes[index].infoPortals.Values);
            ListBoxWorldMaps_SelectedIndexChanged(null, null);
        }

        private void LabelDelMap_Click(object sender, EventArgs e) {
            if(ListBoxWorldMaps.SelectedIndex == -1) return;
            int index = ((Node)ListBoxWorldMaps.SelectedItem).mapNum;
            mapNodes.Remove(index);
            ListBoxWorldMaps.DataSource = new List<Node>(mapNodes.Values);
        }

        private void CheckBoxOtherList_CheckedChanged(object sender, EventArgs e) {
            if(ListBoxWorldMaps.SelectedIndex == -1) return;
            int index = ((Node)ListBoxWorldMaps.SelectedItem).mapNum;

            Node currNode = mapNodes[index];
            currNode.otherList = CheckBoxOtherList.Checked;
            mapNodes[index] = currNode;
        }

        private void ListBoxCurrPortals_PreviewKeyDown(object sender, PreviewKeyDownEventArgs e) { if(e.KeyValue == (int)Keys.Delete) LabelDelPortal_Click(null, null); }
        private void ListBoxWorldMaps_PreviewKeyDown(object sender, PreviewKeyDownEventArgs e) { if(e.KeyValue == (int)Keys.Delete) LabelDelMap_Click(null, null); }
        private void TextBoxMapFolder_KeyPress(object sender, KeyPressEventArgs e) { if(e.KeyChar == (char)Keys.Enter && MapPreviewForm != null && MapPreviewForm.Visible) ListBoxWorldMaps_SelectedIndexChanged(null, null); }
        private void TextBoxWarpsTo_KeyPress(object sender, KeyPressEventArgs e) { if(e.KeyChar == (char)Keys.Enter) BTNAddPortal_Click(null, null); }
        private void TextBoxWarpPortalLocs_KeyPress(object sender, KeyPressEventArgs e) { if(e.KeyChar == (char)Keys.Enter) BTNAddPortal_Click(null, null); }
        private void TextBoxSpecialEndOps_KeyPress(object sender, KeyPressEventArgs e) { if(e.KeyChar == (char)Keys.Enter) BTNAddPortal_Click(null, null); }
        private void TextBoxName_KeyPress(object sender, KeyPressEventArgs e) { if(e.KeyChar == (char)Keys.Enter) ButtonAddMap_Click(null, null); }
        private void TextBoxMapNum_KeyPress(object sender, KeyPressEventArgs e) { if(e.KeyChar == (char)Keys.Enter) ButtonAddMap_Click(null, null); }
        private void TextBoxMapAxis_KeyPress(object sender, KeyPressEventArgs e) { if(e.KeyChar == (char)Keys.Enter) ButtonAddMap_Click(null, null); }
        private void LabelDelMap_MouseEnter(object sender, EventArgs e) { this.Cursor = Cursors.Hand; }
        private void LabelDelMap_MouseLeave(object sender, EventArgs e) { this.Cursor = Cursors.Arrow; }
        private void LabelDelPortal_MouseEnter(object sender, EventArgs e) { this.Cursor = Cursors.Hand; }
        private void LabelDelPortal_MouseLeave(object sender, EventArgs e) { this.Cursor = Cursors.Arrow; }

        private void DAWalkerWorldMapEditorForm_FormClosing(object sender, FormClosingEventArgs e) {
            if(ShowMouseForm != null) ShowMouseForm.Close();
            if(MapPreviewForm != null) MapPreviewForm.Close();
        }

        // DA Process Information
        private void ButtonGetCurrMapNum_Click(object sender, EventArgs e) {
            System.Diagnostics.Process[] myProcessArray = System.Diagnostics.Process.GetProcessesByName("Darkages");

            if(myProcessArray.Length > 0) pReader.OpenProcess((uint)myProcessArray[0].Id);
            else {
                MessageBox.Show("No DA Process Found");
                return;
            }

            TextBoxDAMap.Text = GetDAMap().ToString();
            pReader.Dispose();
        }

        MyMemoryReadSearchLib.MyMemoryReadSearch pReader = new MyMemoryReadSearchLib.MyMemoryReadSearch();
        //Returns -1 if Failed, and Sets Calling macro to false
        private int ReadValues(int lpBaseAddress, uint nSize) {
            byte[] lpBuffer = pReader.ReadProcessMemory((IntPtr)lpBaseAddress, new UIntPtr(nSize));
            if(lpBuffer == null) return -1;
            int ret = lpBuffer[0];

            for(int i = 1; i < nSize; i++) ret += (int)Math.Pow(256, (double)i) * lpBuffer[i];
            return ret;
        }
        private int GetDAMap() {
            int address = ReadValues(DAMapAddr, 4);
            if(address == -1) return -1;
            else return ReadValues(address + DAMapAddrOff, 4);
        }
    }
}