using System;
using System.Collections.Generic;
using System.Drawing;
using System.Windows.Forms;

namespace DAWalkerWorldMapEditor {
    public partial class MapPreview : Form {

        private DAWalkerWorldMapEditorForm MainP;
        private Bitmap DAMapImage;
        private List<ushort[]> myFactors = new List<ushort[]>();
        private MyDAMapLib.MyDAMap myDAMap = new MyDAMapLib.MyDAMap();

        public MapPreview(DAWalkerWorldMapEditorForm MainP) {
            InitializeComponent();
            this.MainP = MainP;
        }

        public void OpenDAMapFile() {
            MapDimList.Items.Clear();

            if((myFactors = MyDAMapLib.MyDAMap.GetFactors(MainP.TextBoxMapFolder.Text + "lod" + MainP.TextBoxMapNum.Text + ".map")) == null) { ClearMap(); return; }
            foreach(ushort[] A in myFactors) MapDimList.Items.Add(A[0] + "x" + A[1]);

            string[] axis = MainP.TextBoxMapAxis.Text.Split(new char[] { ' ' }, StringSplitOptions.RemoveEmptyEntries);
            if(axis.Length == 2 && axis[0] != "-1" && axis[1] != "-1" && MapDimList.Items.Contains(axis[0] + "x" + axis[1])) MapDimList.SelectedItem = axis[0] + "x" + axis[1];
            else MapDimList.SelectedIndex = myFactors.Count / 2;
        }

        private void CreateDABitMap() {
            if(!myDAMap.TryMakeDAMapSquareImage(MainP.TextBoxMapFolder.Text + "lod" + MainP.TextBoxMapNum.Text + ".map", myFactors[MapDimList.SelectedIndex][0], myFactors[MapDimList.SelectedIndex][1], 0.5, Color.Black, System.Drawing.Imaging.PixelFormat.Format16bppRgb555, out DAMapImage)) { ClearMap(); return; }
            this.Width = DAMapImage.Width + 118;
            this.Height = DAMapImage.Height + 43;
            PictureBoxMapPreview.Image = DAMapImage;
        }

        public void ClearMap() {
            myFactors.Clear();
            MapDimList.Items.Clear();
            DAMapImage = null;
            PictureBoxMapPreview.Image = null;
        }

        private void MapDimList_SelectedIndexChanged(object sender, EventArgs e) {
            CreateDABitMap();

            string[] Axis = ((string)MapDimList.SelectedItem).Split(new char[] { 'x' }, StringSplitOptions.RemoveEmptyEntries);
            if(Axis.Length == 2 && (Axis[0] != Axis[1])) MainP.TextBoxMapAxis.Text = Axis[0] + " " + Axis[1];
            else MainP.TextBoxMapAxis.Text = "-1 -1";
        }

        public void DrawSquareBlockAndUpdate(int X, int Y) { DrawSquareBlockAndUpdate(new int[] { X, Y }); }
        public void DrawSquareBlockAndUpdate(int[] portals) { if(MyDAMapLib.MyDAMap.TryColorBlockSquare(portals, Color.Red, 0.5, ref DAMapImage)) PictureBoxMapPreview.Image = DAMapImage; }
        private void PictureBoxMapPreview_MouseClick(object sender, MouseEventArgs e) {
            if(DAMapImage == null) return;

            if(e.Button == MouseButtons.Right) {
                CreateDABitMap();
                MainP.TextBoxWarpPortalLocs.Text = "";
            } else {
                ushort X = (ushort)(e.X / 6), Y = (ushort)(e.Y / 6);
                MainP.TextBoxWarpPortalLocs.Text += " " + X + " " + Y;
                MainP.TextBoxWarpPortalLocs.Text = MainP.TextBoxWarpPortalLocs.Text.Trim();
                DrawSquareBlockAndUpdate(X, Y);
            }
        }
        public void DrawRhombusBlockAndUpdate(int X, int Y) { DrawRhombusBlockAndUpdate(new int[] { X, Y }); }
        public void DrawRhombusBlockAndUpdate(int[] portals) { if(myFactors != null && MyDAMapLib.MyDAMap.TryColorBlockRhombus(portals, myFactors[MapDimList.SelectedIndex][0], myFactors[MapDimList.SelectedIndex][1], Color.Red, 0.5, ref DAMapImage)) PictureBoxMapPreview.Image = DAMapImage; }
        private void PictureBoxMapPreviewRhombus_MouseClick(object sender, MouseEventArgs e) {
            if(DAMapImage == null) return;

            if(e.Button == MouseButtons.Right) {
                CreateDABitMap();
                MainP.TextBoxWarpPortalLocs.Text = "";
            } else {
                ushort blockDimX = (ushort)(10 * 0.5); ushort blockDimY = (ushort)(5 * 0.5);
                ushort mapDimX = myFactors[MapDimList.SelectedIndex][0], mapDimY = myFactors[MapDimList.SelectedIndex][1];

                double X = 1.0 / 2.0 * e.Y / blockDimY + 1.0 / 2.0 * e.X / blockDimX - 1.0 / 2.0 * mapDimY - 1.0 / 2.0;
                double Y = 1.0 / 2.0 * e.Y / blockDimY - 1.0 / 2.0 * e.X / blockDimX + 1.0 / 2.0 * mapDimY + 1.0 / 2.0;

                if(X < 0 || Y < 0 || X > mapDimX || Y > mapDimY) return;

                MainP.TextBoxWarpPortalLocs.Text += " " + (int)X + " " + (int)Y;
                MainP.TextBoxWarpPortalLocs.Text = MainP.TextBoxWarpPortalLocs.Text.Trim();

                DrawRhombusBlockAndUpdate((int)X, (int)Y);
            }
        }

        private void MapPreview_FormClosing(object sender, FormClosingEventArgs e) {
            e.Cancel = true;
            this.Hide();
        }
    }
}